from werkzeug.exceptions import Unauthorized, BadRequest
from sovereign import config, statsd, __version__
from sovereign.dataclasses import DiscoveryRequest
from sovereign.utils.crypto import decrypt, KEY_AVAILABLE, InvalidToken


def validate(auth_string):
    try:
        password = decrypt(auth_string)
    except Exception:
        statsd.increment('discovery.auth.failed')
        raise

    if password in config.passwords:
        statsd.increment('discovery.auth.success')
        return True
    statsd.increment('discovery.auth.failed')
    return False


def authenticate(request: DiscoveryRequest):
    if not config.auth_enabled:
        return
    if not KEY_AVAILABLE:
        raise RuntimeError(
            'No Fernet key loaded, and auth is enabled. '
            'A fernet key must be provided via SOVEREIGN_ENCRYPTION_KEY. '
            'See https://vsyrakis.bitbucket.io/sovereign/docs/html/guides/encryption.html '
            'for more details'
        )
    try:
        encrypted_auth = request.node.metadata['auth']
        assert validate(encrypted_auth)
    except KeyError:
        raise KeyError(f'Discovery request from {request.node.id} is missing auth field')
    except (InvalidToken, AssertionError):
        raise Unauthorized('The authentication provided was invalid')
    except Exception:
        raise BadRequest('The request was malformed')



import gzip
from random import randint
from functools import wraps
from datetime import timedelta
from quart import after_this_request, request, Response
from quart.datastructures import HeaderSet
from cachelib import SimpleCache
from sovereign import statsd


cache = SimpleCache()


def memoize(timeout, jitter=0):
    """
    Decorator to cache a function by name/args

    :param timeout: How long to keep the result
    :param jitter: Randomize the timeout by this many seconds
    """
    if isinstance(timeout, timedelta):
        timeout = timeout.seconds

    timeout += randint(-jitter/2, jitter)

    def decorator(decorated):
        @wraps(decorated)
        def wrapper(*args, **kwargs):
            key = f'{decorated.__name__}{args}{kwargs}'
            ret = cache.get(key)
            metrics_tags = [
                f'function:{decorated.__name__}'
            ]
            if ret is None:
                statsd.increment('cache.miss', tags=metrics_tags)
                ret = decorated(*args, **kwargs)
                cache.set(key, ret, timeout=timeout)
            else:
                statsd.increment('cache.hit', tags=metrics_tags)
            return ret
        return wrapper
    return decorator


def gzcompress(level=2, valid_codes=range(200, 304)):
    """
    Gzip compression decorator from : http://flask.pocoo.org/snippets/122/

    Modified to work with Quart
    Also added configurable compression level & status codes

    Unfortunately does not work with Envoy since it doesn't send gzip related headers
    """
    def decorator(f):
        @wraps(f)
        async def wrapper(*args, **kwargs):
            @after_this_request
            async def compress(response: Response) -> Response:
                accept_encoding = request.headers.get('Accept-Encoding', '').lower()

                if 'gzip' not in accept_encoding:
                    return response

                if response.status_code not in valid_codes or response.content_encoding:
                    return response

                data = await response.get_data()
                output = gzip.compress(data, compresslevel=level)

                response.set_data(output)
                response.content_encoding = 'gzip'
                response.vary = HeaderSet.from_header('Accept-Encoding')
                return response

            return await f(*args, **kwargs)
        return wrapper
    return decorator



import os
import zlib
from dataclasses import dataclass, field
from typing import List
from jinja2 import Template
from sovereign.config_loader import load


@dataclass
class Source:
    type: str
    config: dict


@dataclass
class StatsdConfig:
    host: str = '127.0.0.1'
    port: int = 8125
    tags: dict = field(default_factory=dict)
    namespace: str = 'sovereign'
    enabled: bool = False

    def __post_init__(self):
        # Use config loader to update tags, otherwise they remain a string
        self.tags = {
            k: load(v) for k, v in self.tags.items()
        }


@dataclass
class XdsTemplate:
    path: str
    content: Template = field(init=False)
    checksum: int = field(init=False)

    def __post_init__(self):
        self.content = load(self.path)
        self.checksum = zlib.adler32(self.source.encode())

    @property
    def source(self):
        if 'jinja' in self.path:
            # The Jinja2 template serializer does not properly set a name
            # for the loaded template.
            # The repr for the template prints out as the memory address
            # This makes it really hard to generate a consistent version_info string
            # in rendered configuration.
            # For this reason, we re-load the template as a string instead, and create a checksum.
            path = self.path.replace('+jinja', '+string')
            return load(path)
        else:
            # The only other supported serializers are string, yaml, and json
            # So it should be safe to create this checksum off
            return str(self.content)


@dataclass
class Locality:
    region: str = None
    zone: str = None
    sub_zone: str = None


@dataclass
class Node:
    id: str
    cluster: str
    build_version: str
    locality: Locality = Locality()
    metadata: dict = field(default_factory=dict)

    def __post_init__(self):
        if isinstance(self.locality, dict):
            self.locality = Locality(**self.locality)


@dataclass
class DiscoveryRequest:
    node: Node
    version_info: str = '0'
    resource_names: List[str] = field(default_factory=list)

    def __post_init__(self):
        if isinstance(self.node, dict):
            self.node = Node(**self.node)

    @property
    def envoy_version(self):
        try:
            build_version = self.node.build_version
            revision, version, *other_metadata = build_version.split('/')
        except (AttributeError, ValueError):
            # TODO: log/metric this?
            return 'default'
        return version


@dataclass(frozen=True)
class SovereignConfig:
    sources:                  List[Source]
    templates:                dict
    template_context:         dict = field(default_factory=dict)
    eds_priority_matrix:      dict = field(default_factory=dict)
    modifiers:                List[str] = field(default_factory=list)
    global_modifiers:         List[str] = field(default_factory=list)
    regions:                  List[str] = field(default_factory=list)
    statsd:                   dict = field(default_factory=dict)
    auth_enabled:             bool = bool(os.getenv('SOVEREIGN_AUTH_ENABLED', False))
    auth_passwords:           str = os.getenv('SOVEREIGN_AUTH_PASSWORDS', '')
    encryption_key:           str = os.getenv('SOVEREIGN_ENCRYPTION_KEY', os.getenv('FERNET_ENCRYPTION_KEY'))
    no_changes_response_code: int = int(os.getenv('SOVEREIGN_NO_CHANGE_RESPONSE', 304))
    application_host:         str = os.getenv('SOVEREIGN_HOST', '0.0.0.0')
    application_port:         int = int(os.getenv('SOVEREIGN_PORT', '8080'))
    environment:              str = os.getenv('SOVEREIGN_ENVIRONMENT_TYPE', os.getenv('MICROS_ENVTYPE', 'local'))
    debug_enabled:            bool = bool(os.getenv('SOVEREIGN_DEBUG', False))
    sentry_dsn:               str = os.getenv('SOVEREIGN_SENTRY_DSN')

    @property
    def metrics(self):
        return StatsdConfig(**dict(self.statsd))

    @property
    def passwords(self):
        return self.auth_passwords.split(',') or []
